@@ -47,6 +47,7 @@
 #include "xdp-app-info-snap-private.h"
 #include "xdp-app-info-host-private.h"
 #include "xdp-app-info-test-private.h"
+#include "xdp-enum-types.h"
 #include "xdp-utils.h"
 
 #define DBUS_NAME_DBUS "org.freedesktop.DBus"
@@ -65,16 +66,72 @@ typedef struct _XdpAppInfoPrivate
   char *instance;
   int pidfd;
   GAppInfo *gappinfo;
-  gboolean supports_opath;
-  gboolean has_network;
-  gboolean requires_pid_mapping;
+  XdpAppInfoFlags flags;
 
   /* pid namespace mapping */
   GMutex pidns_lock;
   ino_t pidns_id;
 } XdpAppInfoPrivate;
 
-G_DEFINE_ABSTRACT_TYPE_WITH_PRIVATE (XdpAppInfo, xdp_app_info, G_TYPE_OBJECT)
+static void g_initable_init_iface (GInitableIface *iface);
+
+G_DEFINE_ABSTRACT_TYPE_WITH_CODE (XdpAppInfo, xdp_app_info, G_TYPE_OBJECT,
+                                  G_ADD_PRIVATE (XdpAppInfo)
+                                  G_IMPLEMENT_INTERFACE (G_TYPE_INITABLE, g_initable_init_iface))
+
+enum
+{
+  PROP_0,
+  PROP_ENGINE,
+  PROP_FLAGS,
+  PROP_G_APP_INFO,
+  PROP_ID,
+  PROP_INSTANCE,
+  PROP_PIDFD,
+  N_PROPS
+};
+
+static GParamSpec *properties [N_PROPS];
+
+static gboolean
+xdp_app_info_initable_init (GInitable     *initable,
+                            GCancellable  *cancellable,
+                            GError       **error)
+{
+  XdpAppInfo *app_info = XDP_APP_INFO (initable);
+  XdpAppInfoPrivate *priv = xdp_app_info_get_instance_private (app_info);
+
+  priv->gappinfo =
+    XDP_APP_INFO_GET_CLASS (app_info)->create_gappinfo (app_info);
+
+  if ((priv->flags & XDP_APP_INFO_FLAG_REQUIRE_GAPPINFO) && !priv->gappinfo)
+    {
+      g_set_error (error, G_IO_ERROR, G_IO_ERROR_NOT_FOUND,
+                   "App info not found for '%s'", priv->id);
+      return FALSE;
+    }
+
+  return TRUE;
+}
+
+static void
+g_initable_init_iface (GInitableIface *iface)
+{
+  iface->init = xdp_app_info_initable_init;
+}
+
+static GAppInfo *
+xdp_app_info_real_create_gappinfo (XdpAppInfo *app_info)
+{
+  XdpAppInfoPrivate *priv = xdp_app_info_get_instance_private (app_info);
+  g_autoptr(GAppInfo) gappinfo = NULL;
+  g_autofree char *desktop_id = NULL;
+
+  desktop_id = g_strconcat (priv->id, ".desktop", NULL);
+  gappinfo = G_APP_INFO (g_desktop_app_info_new (desktop_id));
+
+  return g_steal_pointer (&gappinfo);
+}
 
 static void
 xdp_app_info_dispose (GObject *object)
@@ -94,12 +151,140 @@ xdp_app_info_dispose (GObject *object)
   G_OBJECT_CLASS (xdp_app_info_parent_class)->dispose (object);
 }
 
+static void
+xdp_app_info_get_property (GObject    *object,
+                           guint       prop_id,
+                           GValue     *value,
+                           GParamSpec *pspec)
+{
+  XdpAppInfoPrivate *priv =
+    xdp_app_info_get_instance_private (XDP_APP_INFO (object));
+
+  switch (prop_id)
+    {
+    case PROP_ENGINE:
+      g_value_set_string (value, priv->engine);
+      break;
+
+    case PROP_FLAGS:
+      g_value_set_flags (value, priv->flags);
+      break;
+
+    case PROP_G_APP_INFO:
+      g_value_set_object (value, priv->gappinfo);
+      break;
+
+    case PROP_ID:
+      g_value_set_string (value, priv->id);
+      break;
+
+    case PROP_INSTANCE:
+      g_value_set_string (value, priv->instance);
+      break;
+
+    case PROP_PIDFD:
+      g_value_set_int (value, priv->pidfd);
+      break;
+
+    default:
+      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+    }
+}
+
+static void
+xdp_app_info_set_property (GObject      *object,
+                           guint         prop_id,
+                           const GValue *value,
+                           GParamSpec   *pspec)
+{
+  XdpAppInfoPrivate *priv =
+    xdp_app_info_get_instance_private (XDP_APP_INFO (object));
+
+  switch (prop_id)
+    {
+    case PROP_ENGINE:
+      g_assert (priv->engine == NULL);
+      priv->engine = g_value_dup_string (value);
+      break;
+
+    case PROP_FLAGS:
+      g_assert (priv->flags == 0);
+      priv->flags = g_value_get_flags (value);
+      break;
+
+    case PROP_ID:
+      g_assert (priv->id == NULL);
+      priv->id = g_value_dup_string (value);
+      break;
+
+    case PROP_INSTANCE:
+      g_assert (priv->instance == NULL);
+      priv->instance = g_value_dup_string (value);
+      break;
+
+    case PROP_PIDFD:
+      g_assert (priv->pidfd == -1);
+      priv->pidfd = g_value_get_int (value);
+      break;
+
+    default:
+      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+    }
+}
+
 static void
 xdp_app_info_class_init (XdpAppInfoClass *klass)
 {
   GObjectClass *object_class = G_OBJECT_CLASS (klass);
 
   object_class->dispose = xdp_app_info_dispose;
+  object_class->get_property = xdp_app_info_get_property;
+  object_class->set_property = xdp_app_info_set_property;
+
+  klass->create_gappinfo = xdp_app_info_real_create_gappinfo;
+
+  properties[PROP_ENGINE] =
+    g_param_spec_string ("engine", NULL, NULL,
+                         NULL,
+                         G_PARAM_READWRITE |
+                         G_PARAM_CONSTRUCT_ONLY |
+                         G_PARAM_STATIC_STRINGS);
+
+  properties[PROP_FLAGS] =
+    g_param_spec_flags ("flags", NULL, NULL,
+                        XDP_TYPE_APP_INFO_FLAGS,
+                        0,
+                        G_PARAM_READWRITE |
+                        G_PARAM_CONSTRUCT_ONLY |
+                        G_PARAM_STATIC_STRINGS);
+
+  properties[PROP_G_APP_INFO] =
+    g_param_spec_object ("g-app-info", NULL, NULL,
+                         G_TYPE_APP_INFO,
+                         G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);
+
+  properties[PROP_ID] =
+    g_param_spec_string ("id", NULL, NULL,
+                         NULL,
+                         G_PARAM_READWRITE |
+                         G_PARAM_CONSTRUCT_ONLY |
+                         G_PARAM_STATIC_STRINGS);
+
+  properties[PROP_INSTANCE] =
+    g_param_spec_string ("instance", NULL, NULL,
+                         NULL,
+                         G_PARAM_READWRITE |
+                         G_PARAM_CONSTRUCT_ONLY |
+                         G_PARAM_STATIC_STRINGS);
+
+  properties[PROP_PIDFD] =
+    g_param_spec_int ("pidfd", NULL, NULL,
+                      -1, G_MAXINT, -1,
+                      G_PARAM_READWRITE |
+                      G_PARAM_CONSTRUCT_ONLY |
+                      G_PARAM_STATIC_STRINGS);
+
+  g_object_class_install_properties (object_class, N_PROPS, properties);
 }
 
 static void
@@ -110,33 +295,16 @@ xdp_app_info_init (XdpAppInfo *app_info)
   priv->pidfd = -1;
 }
 
-void
-xdp_app_info_initialize (XdpAppInfo *app_info,
-                         const char *engine,
-                         const char *app_id,
-                         const char *instance,
-                         int         pidfd,
-                         GAppInfo   *gappinfo,
-                         gboolean    supports_opath,
-                         gboolean    has_network,
-                         gboolean    requires_pid_mapping)
-{
-  XdpAppInfoPrivate *priv = xdp_app_info_get_instance_private (app_info);
-
-  priv->engine = g_strdup (engine);
-  priv->id = g_strdup (app_id);
-  priv->instance = g_strdup (instance);
-  priv->pidfd = dup (pidfd);
-  g_set_object (&priv->gappinfo, gappinfo);
-  priv->supports_opath = supports_opath;
-  priv->has_network = has_network;
-  priv->requires_pid_mapping = requires_pid_mapping;
-}
-
 gboolean
 xdp_app_info_is_host (XdpAppInfo *app_info)
 {
-  return XDP_IS_APP_INFO_HOST (app_info) ||  XDP_IS_APP_INFO_TEST (app_info);
+  XdpAppInfoPrivate *priv;
+
+  g_return_val_if_fail (app_info != NULL, FALSE);
+
+  priv = xdp_app_info_get_instance_private (app_info);
+
+  return priv->engine == NULL;
 }
 
 const char *
@@ -195,7 +363,7 @@ xdp_app_info_has_network (XdpAppInfo *app_info)
 
   priv = xdp_app_info_get_instance_private (app_info);
 
-  return priv->has_network;
+  return (priv->flags & XDP_APP_INFO_FLAG_HAS_NETWORK) != 0;
 }
 
 gboolean
@@ -234,7 +402,6 @@ static char *
 remap_path (XdpAppInfo *app_info,
             const char *path)
 {
-
   if (!XDP_APP_INFO_GET_CLASS (app_info)->remap_path)
     return g_strdup (path);
 
@@ -262,8 +429,9 @@ verify_proc_self_fd (XdpAppInfo  *app_info,
   path_buffer[symlink_size] = 0;
 
   /* All normal paths start with /, but some weird things
-     don't, such as socket:[27345] or anon_inode:[eventfd].
-     We don't support any of these */
+   * don't, such as socket:[27345] or anon_inode:[eventfd].
+   * We don't support any of these.
+   */
   if (path_buffer[0] != '/')
     {
       g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_FILENAME,
@@ -272,11 +440,12 @@ verify_proc_self_fd (XdpAppInfo  *app_info,
     }
 
   /* File descriptors to actually deleted files have " (deleted)"
-     appended to them. This also happens to some fake fd types
-     like shmem which are "/<name> (deleted)". All such
-     files are considered invalid. Unfortunately this also
-     matches files with filenames that actually end in " (deleted)",
-     but there is not much to do about this. */
+   * appended to them. This also happens to some fake fd types
+   * like shmem which are "/<name> (deleted)". All such
+   * files are considered invalid. Unfortunately this also
+   * matches files with filenames that actually end in " (deleted)",
+   * but there is not much to do about this.
+   */
   if (g_str_has_suffix (path_buffer, " (deleted)"))
     {
       const char *mountpoint = xdp_get_documents_mountpoint ();
@@ -284,11 +453,12 @@ verify_proc_self_fd (XdpAppInfo  *app_info,
       if (mountpoint != NULL && g_str_has_prefix (path_buffer, mountpoint))
         {
           /* Unfortunately our workaround for dcache purging triggers
-             o_path file descriptors on the fuse filesystem being
-             marked as deleted, so we have to allow these here and
-             rewrite them. This is safe, becase we will stat the file
-             and compare to make sure we end up on the right file. */
-          path_buffer[symlink_size - strlen(" (deleted)")] = 0;
+           * o_path file descriptors on the fuse filesystem being
+           * marked as deleted, so we have to allow these here and
+           * rewrite them. This is safe, becase we will stat the file
+           * and compare to make sure we end up on the right file.
+           */
+          path_buffer[symlink_size - strlen (" (deleted)")] = 0;
         }
       else
         {
@@ -438,7 +608,7 @@ xdp_app_info_get_path_for_fd (XdpAppInfo   *app_info,
           return NULL;
         }
 
-      if (!priv->supports_opath)
+      if ((priv->flags & XDP_APP_INFO_FLAG_SUPPORTS_OPATH) == 0)
         {
           g_set_error (error, G_IO_ERROR, G_IO_ERROR_NOT_SUPPORTED,
                        "App \"%s\" of type %s does not support O_PATH fd passing",
@@ -611,9 +781,7 @@ xdp_connection_get_pidfd (GDBusConnection  *connection,
   uint32_t pid;
   int fd_index;
   g_autoptr(GUnixFDList) fd_list = NULL;
-  int fds_len = 0;
-  const int *fds;
-  int pidfd;
+  g_autofd int pidfd = -1;
 
   reply = g_dbus_connection_call_with_unix_fd_list_sync (connection,
                                                          DBUS_NAME_DBUS,
@@ -676,21 +844,17 @@ xdp_connection_get_pidfd (GDBusConnection  *connection,
       return FALSE;
     }
 
-  fds = g_unix_fd_list_peek_fds (fd_list, &fds_len);
-  if (fds_len <= fd_index)
+  if (fd_index >= g_unix_fd_list_get_length (fd_list))
     {
-      g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED, "Can't find peer pidfd");
+      g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED, "Pidfd index is out of bounds");
       return FALSE;
     }
 
-  pidfd = dup (fds[fd_index]);
+  pidfd = g_unix_fd_list_get (fd_list, fd_index, error);
   if (pidfd < 0)
-    {
-      g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED, "Can't dup pidfd");
-      return FALSE;
-    }
+    return FALSE;
 
-  *out_pidfd = pidfd;
+  *out_pidfd = g_steal_fd (&pidfd);
   *out_pid = pid;
   return TRUE;
 }
@@ -805,7 +969,7 @@ xdp_connection_create_app_info_sync (GDBusConnection  *connection,
 
   if (app_info == NULL)
     {
-      app_info = xdp_app_info_flatpak_new (pid, pidfd, &local_error);
+      app_info = xdp_app_info_flatpak_new (pid, &pidfd, &local_error);
       if (app_info)
         app_info_kind = "flatpak";
     }
@@ -820,7 +984,7 @@ xdp_connection_create_app_info_sync (GDBusConnection  *connection,
 
   if (app_info == NULL)
     {
-      app_info = xdp_app_info_snap_new (pid, pidfd, &local_error);
+      app_info = xdp_app_info_snap_new (pid, &pidfd, &local_error);
       if (app_info)
         app_info_kind = "snap";
     }
@@ -835,11 +999,11 @@ xdp_connection_create_app_info_sync (GDBusConnection  *connection,
 
   if (app_info == NULL)
     {
-      app_info = xdp_app_info_host_new (pid, pidfd);
+      app_info = xdp_app_info_host_new (pid, &pidfd);
       app_info_kind = "derived host";
     }
 
-  g_return_val_if_fail (app_info != NULL, NULL);
+  g_assert (XDP_IS_APP_INFO (app_info));
 
   g_debug ("Adding %s app '%s'", app_info_kind, xdp_app_info_get_id (app_info));
 
@@ -876,7 +1040,7 @@ xdp_connection_create_host_app_info_sync (GDBusConnection  *connection,
       if (is_sandboxed)
         {
           g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
-                       "Can't manually register a flatpak application");
+                       "Can't manually register a Flatpak application");
           return NULL;
         }
 
@@ -886,11 +1050,11 @@ xdp_connection_create_host_app_info_sync (GDBusConnection  *connection,
       if (is_sandboxed)
         {
           g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
-                       "Can't manually register a snap application");
+                       "Can't manually register a Snap application");
           return NULL;
         }
 
-      app_info = xdp_app_info_host_new_registered (pidfd, app_id, error);
+      app_info = xdp_app_info_host_new_registered (&pidfd, app_id, error);
       if (!app_info)
         return NULL;
     }
